from math import inf
from colibrie.geometry import Point, Rect
import numpy as np


def is_table_rotated(data):
    total_span = 0
    rotated_span = 0
    for row in data:
        for column in row:
            if column:
                for span in column:
                    total_span += 1
                    # print(span['dir'])
                    if span["dir"] != (1, 0):
                        rotated_span += 1
            # column = column_text

    # all span are rotated inside the table
    if (rotated_span / total_span) == 1:
        return True

    return False


def extract_text_from_spans(spans):
    text = ""
    for span in spans:
        text += (
            span["text"] + "<br>\n"
        )  # TODO REMOVE <br> which is here for visual purpose now

    return text


class RangeMap:
    def __init__(self, dictionary, minimum=-inf, maximum=inf):
        assert dictionary
        self._dictionary = dictionary
        self.min, self.max = minimum, maximum
        self._generate_tree()

    def __delitem__(self, key):
        del self._dictionary[key]
        self._generate_tree()

    def __getitem__(self, item):
        current = self._tree
        while current:
            if item < current.lower:
                current = current.before
            elif item <= current.upper:
                return current.value
            else:
                current = current.after
        raise Exception("Tree not constructed properly")

    def __setitem__(self, key, value):
        self._dictionary[key] = value
        self._generate_tree()

    def _generate_tree(self):
        range_mapping = [
            (((p + c) / 2, (c + n) / 2), self._dictionary[c])
            for p, c, n in self._iterate_keys()
        ]
        self._tree = RangeMap.Node(range_mapping)

    def _iterate_keys(self):
        keys = sorted(self._dictionary.keys())
        return zip([self.min] + keys[:-1], keys, keys[1:] + [self.max])

    class Node:
        def __init__(self, range_mapping):
            assert range_mapping
            middle = len(range_mapping) // 2
            (self.lower, self.upper), self.value = range_mapping[middle]
            before, after = range_mapping[:middle], range_mapping[middle + 1 :]
            self.before = RangeMap.Node(before) if before else None
            self.after = RangeMap.Node(after) if after else None


def intersect(p1, p2, p3, p4):
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x4, y4 = p4
    denom = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1)
    if denom == 0:  # parallel
        return None
    ua = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / denom
    if ua < 0 or ua > 1:  # out of range
        return None
    ub = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / denom
    if ub < 0 or ub > 1:  # out of range
        return None
    x = x1 + ua * (x2 - x1)
    y = y1 + ua * (y2 - y1)
    return Point(x, y)


def adjust_line_lenght(lines, direction, revert=False):
    adjusted_lines = []

    if direction == "vertical":
        for line in lines:
            point_a = line[0]
            point_b = line[1]
            point_a.y -= 3 if not revert else -3
            point_b.y += 3 if not revert else -3

            adjusted_lines.append((point_a, point_b))
    elif direction == "horizontal":
        for line in lines:
            point_a = line[0]
            point_b = line[1]
            point_a.x -= 3 if not revert else -3
            point_b.x += 3 if not revert else -3

            adjusted_lines.append((point_a, point_b))
    else:
        return adjusted_lines

    return adjusted_lines


def is_rectangle_overlap(r1, r2):
    if (r1[0] >= r2[2]) or (r1[2] <= r2[0]) or (r1[3] <= r2[1]) or (r1[1] >= r2[3]):
        return False
    else:
        return True


def is_rectangle_contained(r1, r2):
    return r1[0] <= r2[0] <= r2[2] <= r1[2] and r1[1] <= r2[1] <= r2[3] <= r1[3]


def get_text_in_cell(rect, text):
    text_in_cell = []

    # Give a 1 px text tolerance around the rect
    rect = Rect(rect.x0 - 1, rect.y0 - 1, rect.x1 + 1, rect.y1 + 1)

    # skip block / lines that are not relevant because no intersection with the cell Rect to speed up execution times
    # by 10%
    for block in text.get("blocks"):
        bbox = block["bbox"]
        if is_rectangle_overlap(
            [bbox[0], bbox[1], bbox[2], bbox[3]], [rect.x0, rect.y0, rect.x1, rect.y1]
        ):

            for line in block.get("lines"):
                bbox = line["bbox"]
                if is_rectangle_overlap(
                    [bbox[0], bbox[1], bbox[2], bbox[3]],
                    [rect.x0, rect.y0, rect.x1, rect.y1],
                ):

                    for span in line.get("spans"):
                        bbox = span["bbox"]

                        if is_rectangle_contained(
                            [rect.x0, rect.y0, rect.x1, rect.y1],
                            [bbox[0], bbox[1], bbox[2], bbox[3]],
                        ):
                            span["dir"] = line["dir"]
                            text_in_cell.append(span)

    return text_in_cell


def closest_value(input_list, input_value):
    arr = np.asarray(input_list)

    i = (np.abs(arr - input_value)).argmin()

    return arr[i]



from colibrie.utils import (
    intersect,
    adjust_line_lenght,
)


def get_intersection_between_horizontal_and_vertical_lines(
    horizontal_lines, vertical_lines
):
    intersections = {}

    vertical_lines = adjust_line_lenght(vertical_lines, "vertical")
    horizontal_lines = adjust_line_lenght(horizontal_lines, "horizontal")

    for horizontal_line in horizontal_lines:
        for vertical_line in vertical_lines:
            point_a, point_b = horizontal_line
            point_c, point_d = vertical_line
            intersection_point = intersect(point_a, point_b, point_c, point_d)
            if intersection_point:
                intersections[intersection_point] = [vertical_line, horizontal_line]

    vertical_lines = adjust_line_lenght(vertical_lines, "vertical", revert=True)
    horizontal_lines = adjust_line_lenght(horizontal_lines, "horizontal", revert=True)

    return intersections



class Rect:
    """Rect() - all zeros
    Rect(x0, y0, x1, y1) - 4 coordinates
    """

    def __init__(self, x0, y0, x1, y1):
        self.x0, self.y0, self.x1, self.y1 = x0, y0, x1, y1
        return None

    def __getitem__(self, i):
        return (self.x0, self.y0, self.x1, self.y1)[i]

    def __len__(self):
        return 4

    def __repr__(self):
        return "Rect" + str(tuple(self))


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __getitem__(self, i):
        return (self.x, self.y)[i]

    def __len__(self):
        return 2

    def __setitem__(self, i, v):
        v = float(v)
        if i == 0:
            self.x = v
        elif i == 1:
            self.y = v
        else:
            raise IndexError("index out of range")
        return None

    def __repr__(self):
        return "Point" + str(tuple(self))

    def __pos__(self):
        return Point(self)

    def __neg__(self):
        return Point(-self.x, -self.y)

    def __bool__(self):
        return not (max(self) == min(self) == 0)

    def __nonzero__(self):
        return not (max(self) == min(self) == 0)

    def __eq__(self, p):
        if not hasattr(p, "__len__"):
            return False
        return len(p) == 2 and bool(self - p) is False

    def __sub__(self, p):
        if hasattr(p, "__float__"):
            return Point(self.x - p, self.y - p)
        if len(p) != 2:
            raise ValueError("Point: bad seq len")
        return Point(self.x - p[0], self.y - p[1])

    def __hash__(self):
        return hash(tuple(self))



import math

from ailist import AIList

from colibrie.utils import RangeMap, closest_value

from colibrie.geometry import Rect, Point


def get_lines_fragmented(page):
    lines_fragmented = []

    drawings = page.get_cdrawings()

    for drawing in drawings:
        # If there is a curve in the drawing, we considere it not valid for a rectangle
        for part in drawing["items"]:
            if part[0] == "c":
                break
        else:
            # By default if the key closePath is not present, it's a segment
            # so it's not a closePath
            if drawing.get("closePath", False):
                continue

            rect = drawing["rect"]
            rect = Rect(rect[0], rect[1], rect[2], rect[3])
            if math.isclose(rect.x0, rect.x1, abs_tol=3):
                rect.x1 = rect.x0
            if math.isclose(rect.y0, rect.y1, abs_tol=3):
                rect.y1 = rect.y0

            rect.x0 = round(rect.x0) if rect.x0 > 0 else 0
            rect.x1 = round(rect.x1) if rect.x1 > 0 else 0
            rect.y0 = round(rect.y0) if rect.y0 > 0 else 0
            rect.y1 = round(rect.y1) if rect.y1 > 0 else 0

            upper_line = (Point(rect.x0, rect.y0), Point(rect.x1, rect.y0))
            lower_line = (Point(rect.x0, rect.y1), Point(rect.x1, rect.y1))
            left_line = (Point(rect.x0, rect.y0), Point(rect.x0, rect.y1))
            right_line = (Point(rect.x1, rect.y0), Point(rect.x1, rect.y1))

            lines_fragmented += [upper_line, lower_line, left_line, right_line]

    return lines_fragmented


def get_vertical_fragmented_lines(lines_fragmented):
    return [
        line
        for line in lines_fragmented
        if math.isclose(line[0].x, line[1].x, abs_tol=5) and line[0].y != line[1].y
    ]


def get_horizontal_fragmented_lines(lines_fragmented):
    return [
        line
        for line in lines_fragmented
        if math.isclose(line[0].y, line[1].y, abs_tol=5) and line[0].x != line[1].x
    ]


def generate_missing_horizontal_lines(horizontal_lines, vertical_lines, distinct_y_lst):
    if not horizontal_lines or not vertical_lines:
        return horizontal_lines

    y_aligned_point = {y: [] for y in distinct_y_lst}

    if not y_aligned_point:
        return horizontal_lines

    range_map = RangeMap(y_aligned_point)
    for l in vertical_lines:
        for point in l:
            range_map[point.y].append(point)

    for y, points in range_map._dictionary.items():
        if len(points) > 1:
            point_a = Point(points[0].x, points[0].y)
            point_b = Point(points[-1].x, points[-1].y)
            horizontal_lines.append((point_a, point_b))

    horizontal_lines = remove_overlaping_horizontal_lines(horizontal_lines)

    return horizontal_lines


def generate_missing_vertical_lines(horizontal_lines, vertical_lines, distinct_x_lst):
    if not horizontal_lines or not vertical_lines:
        return vertical_lines

    x_aligned_point = {x: [] for x in distinct_x_lst}

    if not x_aligned_point:
        return vertical_lines

    range_map = RangeMap(x_aligned_point)
    for l in horizontal_lines:
        for point in l:
            range_map[point.x].append(point)

    for x, points in range_map._dictionary.items():
        if len(points) > 1:
            point_a = Point(points[0].x, points[0].y)
            point_b = Point(points[-1].x, points[-1].y)
            vertical_lines.append((point_a, point_b))

    vertical_lines = remove_overlaping_vertical_lines(vertical_lines)

    return vertical_lines


def remove_overlaping_vertical_lines(vertical_lines):
    merged_vertical_lines = []

    vertical_lines.sort(key=lambda line: line[1].x)

    group = []
    current_x = None
    for line in vertical_lines:
        if not current_x:
            current_x = line[0].x
            group.append(line)
            continue

        if line[0].x != current_x:
            interval_tree = AIList()
            [interval_tree.add(line[0].y, line[1].y) for line in group]

            merged_interval = interval_tree.merge(gap=5)

            merged_vertical_lines += [
                (Point(current_x, interval.start), Point(current_x, interval.end))
                for interval in merged_interval
            ]
            group = [line]
            current_x = line[0].x
        else:
            group.append(line)

    if group:
        interval_tree = AIList()
        [interval_tree.add(line[0].y, line[1].y) for line in group]
        merged_interval = interval_tree.merge(gap=5)
        merged_vertical_lines += [
            (Point(current_x, interval.start), Point(current_x, interval.end))
            for interval in merged_interval
        ]

    return merged_vertical_lines


def remove_overlaping_horizontal_lines(horizontal_lines):
    merged_horizontal_lines = []

    horizontal_lines.sort(key=lambda line: (line[1].y))

    group = []
    current_y = None
    for line in horizontal_lines:
        if not current_y:
            current_y = line[0].y
            group.append(line)
            continue

        if line[0].y != current_y:
            interval_tree = AIList()
            [interval_tree.add(line[0].x, line[1].x) for line in group]

            merged_interval = interval_tree.merge(gap=5)

            merged_horizontal_lines += [
                (Point(interval.start, current_y), Point(interval.end, current_y))
                for interval in merged_interval
            ]
            group = [line]
            current_y = line[0].y
        else:
            group.append(line)

    if group:
        interval_tree = AIList()
        [interval_tree.add(line[0].x, line[1].x) for line in group]
        merged_interval = interval_tree.merge(gap=5)
        merged_horizontal_lines += [
            (Point(interval.start, current_y), Point(interval.end, current_y))
            for interval in merged_interval
        ]

    return merged_horizontal_lines


def normalize_vertical_lines(vertical_lines, horizontal_lines):
    """
    This function will normalize coordinate between horizontal and vertical lines
    like if a vertical line that in y = 87 and the horizontal lines corresponding has
    a y = 86 then the vertical line will have this y value set to 86 by calculating
    the closest Y value from horizontal lines
    """
    distinct_horizontal_y = list(
        set([point.y for lines in horizontal_lines for point in lines])
    )

    for vertical_line in vertical_lines:
        closest_y0 = closest_value(distinct_horizontal_y, vertical_line[0].y)
        closest_y1 = closest_value(distinct_horizontal_y, vertical_line[1].y)
        vertical_line[0].y = (
            closest_y0
            if math.isclose(vertical_line[0].y, closest_y0, abs_tol=3)
            else vertical_line[0].y
        )
        vertical_line[1].y = (
            closest_y1
            if math.isclose(vertical_line[1].y, closest_y1, abs_tol=3)
            else vertical_line[1].y
        )

    return vertical_lines


def normalize_horizontal_lines(vertical_lines, horizontal_lines):
    """
    This function will normalize coordinate between horizontal and vertical lines
    like if a vertical line that in y = 87 and the horizontal lines corresponding has
    a y = 86 then the vertical line will have this y value set to 86 by calculating
    the closest Y value from horizontal lines
    """
    distinct_vertical_x = list(
        set([point.x for lines in vertical_lines for point in lines])
    )

    for horizontal_line in horizontal_lines:
        closest_x0 = closest_value(distinct_vertical_x, horizontal_line[0].x)
        closest_x1 = closest_value(distinct_vertical_x, horizontal_line[1].x)
        horizontal_line[0].x = (
            closest_x0
            if math.isclose(horizontal_line[0].x, closest_x0, abs_tol=3)
            else horizontal_line[0].x
        )
        horizontal_line[1].x = (
            closest_x1
            if math.isclose(horizontal_line[1].x, closest_x1, abs_tol=3)
            else horizontal_line[1].x
        )

    return horizontal_lines



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['colibrie']

package_data = \
{'': ['*']}

install_requires = \
['Cython>=0.29.32,<0.30.0',
 'Pillow>=9.2.0,<10.0.0',
 'PyMuPDF>=1.20.2,<2.0.0',
 'ailist>=1.0.4,<2.0.0',
 'numpy>=1.23.3,<2.0.0',
 'pandas>=1.5.0,<2.0.0']

setup_kwargs = {
    'name': 'colibrie',
    'version': '1.1.2',
    'description': 'Colibrie is a blazing fast repository to extract tables from PDF files',
    'long_description': "# Colibrie\nColibrie is a blazing fast repository to extract tables from PDF files \n\n# Installation\n### installation using source\n```\npip install poetry\n\ngit clone https://github.com/abitoun-42/colibrie.git\n\ncd colibrie\n\npoetry install\n```\n### installation using pip\n```\npip install colibrie\n```\n\n# Usage\n```\nfrom colibrie.extract_tables import extract_table\n\ntables = extract_table('pdf_path')\n```",
    'author': 'Alexandre Bitoun',
    'author_email': 'alexandre.bitoun@outlook.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/abitoun-42/colibrie',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import fitz

from colibrie.tables import create_table, process_table, get_tables_candidates

from colibrie.intersection import get_intersection_between_horizontal_and_vertical_lines

from colibrie.lines import (
    get_lines_fragmented,
    get_horizontal_fragmented_lines,
    get_vertical_fragmented_lines,
    remove_overlaping_horizontal_lines,
    remove_overlaping_vertical_lines,
    generate_missing_horizontal_lines,
    generate_missing_vertical_lines,
    normalize_vertical_lines,
    normalize_horizontal_lines,
)


def extract_tables(filepath, debug_mode=False, preserve_span=True):
    """
    :param: preserve_span: bool
        give a possibility to speed_up the execution
        by not preserving information about text position and only keeping
        the text content, can be useful for those with proper table aligned that
        do not need to do extra work with it
    """

    doc = fitz.Document(filepath)

    tables_lst = []

    for page in doc:

        ####################### GET LINE ######################################

        lines_fragmented = get_lines_fragmented(page)

        vertical_lines = get_vertical_fragmented_lines(lines_fragmented)

        horizontal_lines = get_horizontal_fragmented_lines(lines_fragmented)

        #######################################################################

        ################## REMOVE OVERLAPING LINES ############################

        vertical_lines = remove_overlaping_vertical_lines(vertical_lines)

        horizontal_lines = remove_overlaping_horizontal_lines(horizontal_lines)

        #######################################################################

        if not vertical_lines or not horizontal_lines:
            continue

        #################### NORMALIZE Y X COORDINATE ############################

        vertical_lines = normalize_vertical_lines(vertical_lines, horizontal_lines)

        horizontal_lines = normalize_horizontal_lines(vertical_lines, horizontal_lines)

        ########################################################################

        #################### GET TABLES CANDIDATES #############################
        tables_candidates = get_tables_candidates(vertical_lines, horizontal_lines)
        ########################################################################

        for table_candidate in tables_candidates:

            vertical_lines = get_vertical_fragmented_lines(table_candidate)

            horizontal_lines = get_horizontal_fragmented_lines(table_candidate)

            ########################## CREATE UNEXISTING MISSING LINE ###########

            distinct_y_lst = set(
                [point.y for lines in vertical_lines for point in lines]
            )
            distinct_x_lst = set(
                [point.x for lines in horizontal_lines for point in lines]
            )

            if not distinct_x_lst or not distinct_y_lst:
                continue

            range_y = (min(distinct_y_lst), max(distinct_y_lst))
            range_x = (min(distinct_x_lst), max(distinct_x_lst))

            # If the table does not meet the condition of a minimum size
            # we assume it is a bad table
            if (range_y[1] - range_y[0]) <= 20 and (range_x[1] - range_x[0]) <= 50:
                continue

            horizontal_lines = generate_missing_horizontal_lines(
                horizontal_lines, vertical_lines, distinct_y_lst
            )

            vertical_lines = generate_missing_vertical_lines(
                horizontal_lines, vertical_lines, distinct_x_lst
            )

            ########################################################################

            ################### GET INTERSECTIONS #################################

            intersections = get_intersection_between_horizontal_and_vertical_lines(
                horizontal_lines, vertical_lines
            )

            if not len(intersections) >= 6:
                continue

            #########################################################################

            ###################### CREATE TABLE LIST ################################
            table = create_table(intersections, horizontal_lines, vertical_lines)

            #########################################################################

            #################################FILL TABLE##############################

            table = process_table(page, table, preserve_span, intersections)

            if not table:
                continue

            #########################################################################

            if debug_mode:
                pix = page.get_pixmap()
                table.debug.image = pix.tobytes()

            tables_lst.append(table)

    doc.close()

    return tables_lst
